@@ -81,12 +81,20 @@ class mergestate(object):
             self.mark(dfile, 'r')
         return r
 
-def _checkunknown(wctx, mctx):
+def _checkunknown(wctx, mctx, folding):
     "check for collisions between unknown files and files in mctx"
-    for f in wctx.unknown():
-        if f in mctx and mctx[f].cmp(wctx[f]):
+    if folding:
+        foldf = util.normcase
+    else:
+        foldf = lambda fn: fn
+    folded = {}
+    for fn in mctx:
+        folded[foldf(fn)] = fn
+    for fn in wctx.unknown():
+        f = foldf(fn)
+        if f in folded and mctx[folded[f]].cmp(wctx[f]):
             raise util.Abort(_("untracked file in working directory differs"
-                               " from file in requested revision: '%s'") % f)
+                               " from file in requested revision: '%s'") % fn)
 
 def _checkcollision(mctx):
     "check for case folding collisions in the destination context"
@@ -273,7 +281,6 @@ def applyupdates(repo, action, wctx, mctx, actx, overwrite):
     action.sort(key=actionkey)
 
     # prescan for merges
-    u = repo.ui
     for a in action:
         f, m = a[:2]
         if m == 'm': # merge
@@ -308,8 +315,8 @@ def applyupdates(repo, action, wctx, mctx, actx, overwrite):
     numupdates = len(action)
     for i, a in enumerate(action):
         f, m = a[:2]
-        u.progress(_('updating'), i + 1, item=f, total=numupdates,
-                   unit=_('files'))
+        repo.ui.progress(_('updating'), i + 1, item=f, total=numupdates,
+                         unit=_('files'))
         if f and f[0] == "/":
             continue
         if m == "r": # remove
@@ -377,7 +384,7 @@ def applyupdates(repo, action, wctx, mctx, actx, overwrite):
             repo.wopener.audit(f)
             util.setflags(repo.wjoin(f), 'l' in flags, 'x' in flags)
     ms.commit()
-    u.progress(_('updating'), None, total=numupdates, unit=_('files'))
+    repo.ui.progress(_('updating'), None, total=numupdates, unit=_('files'))
 
     return updated, merged, removed, unresolved
 
@@ -538,9 +545,10 @@ def update(repo, node, branchmerge, force, partial, ancestor=None):
         ### calculate phase
         action = []
         wc.status(unknown=True) # prime cache
+        folding = not util.checkcase(repo.path)
         if not force:
-            _checkunknown(wc, p2)
-        if not util.checkcase(repo.path):
+            _checkunknown(wc, p2, folding)
+        if folding:
             _checkcollision(p2)
         action += _forgetremoved(wc, p2, branchmerge)
         action += manifestmerge(repo, wc, p2, pa, overwrite, partial)
